/* 
 * Copyright (C) 2004, 2005 Jean-Yves Lefort <jylefort@brutele.be>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include <string.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <translate.h>
#include <enchant.h>
#include <gconf/gconf-client.h>
#include "gt-language-detection.h"
#include "gt-app.h"
#include "gt-util.h"
#include "gt-conf.h"
#include "gt-shell.h"

typedef struct
{
  int			ref_count;
  char			*tag;
  EnchantDict   *speller;
  int			score;
} Speller;

typedef enum
{
  REQUEST_DETECT_LANGUAGE,
  REQUEST_DISABLE
} RequestType;

typedef struct
{
  RequestType		type;
  GSList		*spellers;
  GSList		*words;
} Request;

G_LOCK_DEFINE_STATIC(self);

static EnchantBroker *broker = NULL;
static GSList *spellers = NULL;
static GAsyncQueue *detect_queue = NULL;

static GTLanguageDetectionResultFunc result_func = NULL;
static gpointer result_user_data;

static gboolean autodetect_language;

static void gt_language_detection_notify_pairs_h (GObject *object,
						  GParamSpec *pspec,
						  gpointer user_data);

static void gt_language_detection_create_spellers (void);
static gpointer gt_language_detection_create_spellers_thread (gpointer data);

static void gt_language_detection_autodetect_language_notify_cb (GConfClient *client,
								 unsigned int cnxn_id,
								 GConfEntry *entry,
								 gpointer user_data);
static void gt_language_detection_enable (void);
static void gt_language_detection_disable (void);

static Speller *gt_language_detection_speller_new (const char *tag);
static Speller *gt_language_detection_speller_ref (Speller *speller);
static void gt_language_detection_speller_unref (Speller *speller);

static Request *gt_language_detection_request_new (RequestType type);
static void gt_language_detection_request_free (Request *request);

static gpointer gt_language_detection_detect_thread (gpointer data);
static gboolean gt_language_detection_use_word (const char *word);
static int gt_language_detection_speller_compare_tag (const Speller *a, const char *tag);
static int gt_language_detection_speller_compare_score (const Speller *a, const Speller *b);

/*** implementation **********************************************************/

void
gt_language_detection_init (GTLanguageDetectionResultFunc func, gpointer user_data)
{
  GConfClient *client;

  g_return_if_fail(result_func == NULL);
  g_return_if_fail(func != NULL);

  result_func = func;
  result_user_data = user_data;

  client = gconf_client_get_default ();

  autodetect_language = gconf_client_get_bool(client, GT_CONF_AUTODETECT_LANGUAGE, NULL);

  if (autodetect_language)
    gt_language_detection_enable();

  gconf_client_notify_add(client, GT_CONF_AUTODETECT_LANGUAGE, gt_language_detection_autodetect_language_notify_cb, NULL, NULL, NULL);
  g_signal_connect(gt_shell_get_translate_session(gt_shell), "notify::pairs", G_CALLBACK(gt_language_detection_notify_pairs_h), NULL);
  g_object_unref (client);
}

static void
gt_language_detection_notify_pairs_h (GObject *object,
				      GParamSpec *pspec,
				      gpointer user_data)
{
  if (autodetect_language)
    gt_language_detection_create_spellers();
}

static void
gt_language_detection_create_spellers (void)
{
  TranslateSession *session;

  if (!broker)
    broker = enchant_broker_init();

  session = gt_shell_get_translate_session(gt_shell);
  gt_thread_create(gt_app_window, gt_language_detection_create_spellers_thread, g_object_ref(session));
}

static gpointer
gt_language_detection_create_spellers_thread (gpointer data)
{
  TranslateSession *session = data;
  GSList *next;
  GSList *pairs;
  GSList *new_spellers = NULL;
  GHashTable *from_targets;

  from_targets = g_hash_table_new(g_str_hash, g_str_equal);

  pairs = translate_session_get_pairs(session);

  for (next = g_slist_next(pairs); next != NULL; next = g_slist_next (next))
    {
      TranslatePair *pair = next->data;
      const char *lang_code;

      lang_code = pair ? translate_pair_get_from(pair) : NULL;
	  
      if (lang_code
	  && (translate_pair_get_flags(pair) & TRANSLATE_PAIR_TEXT)
	  && ! g_hash_table_lookup(from_targets, lang_code))
	{
	  Speller *speller = gt_language_detection_speller_new(lang_code);
	  if (speller)
	    new_spellers = g_slist_append(new_spellers, speller);
          g_hash_table_insert(from_targets, (gpointer) lang_code, (gpointer)1);
	}
    }

  g_hash_table_destroy(from_targets);
  gt_g_object_slist_free(pairs);

  G_LOCK(self);

  g_slist_foreach(spellers, (GFunc) gt_language_detection_speller_unref, NULL);
  g_slist_free(spellers);
  spellers = new_spellers;

  if (g_slist_length(spellers) >= 2)
    gt_language_detection_enable();
  else
    {
      GDK_THREADS_ENTER();
      gt_error_dialog(gt_app_window,
		      _("Language detection not available"),
		      _("Less than two languages or spelling dictionaries are available."));
      gdk_flush();
      GDK_THREADS_LEAVE();

      gt_language_detection_disable();
    }

  G_UNLOCK(self);

  g_object_unref(session);
  return NULL;
}

static void
gt_language_detection_autodetect_language_notify_cb (GConfClient *client,
						     unsigned int cnxn_id,
						     GConfEntry *entry,
						     gpointer user_data)
{
  const GConfValue *value = gconf_entry_get_value(entry);

  G_LOCK(self);

  autodetect_language = gconf_value_get_bool(value);
  if (autodetect_language)
    gt_language_detection_enable();
  else
    gt_language_detection_disable();

  G_UNLOCK(self);
}

static void
gt_language_detection_enable (void)
{
  if (! detect_queue)
    {
      detect_queue = g_async_queue_new();
      gt_language_detection_create_spellers();
      gt_thread_create(gt_app_window, gt_language_detection_detect_thread, detect_queue);
    }
}

static void
gt_language_detection_disable (void)
{
  if (detect_queue)
    {
      g_async_queue_push(detect_queue, gt_language_detection_request_new(REQUEST_DISABLE));
      detect_queue = NULL;	/* unreffed by thread */
      g_slist_foreach(spellers, (GFunc) gt_language_detection_speller_unref, NULL);
      g_slist_free(spellers);
      spellers = NULL;
      if (broker)
        {
          enchant_broker_free(broker);
          broker = NULL;
        }
    }
}

static Speller *
gt_language_detection_speller_new (const char *tag)
{
  Speller *speller = NULL;
  EnchantDict *dict;

  if ((dict = enchant_broker_request_dict(broker, tag)))
	{
	  speller = g_new0(Speller, 1);
	  speller->ref_count = 1;
	  speller->tag = g_strdup(tag);
      speller->speller = dict;
    }
  
  return speller;
}

static Speller *
gt_language_detection_speller_ref (Speller *speller)
{
  g_return_val_if_fail(speller != NULL, NULL);

  g_atomic_int_inc(&speller->ref_count);

  return speller;
}

static void
gt_language_detection_speller_unref (Speller *speller)
{
  g_return_if_fail(speller != NULL);

  if (g_atomic_int_dec_and_test(&speller->ref_count))
    {
      g_free(speller->tag);
      enchant_broker_free_dict(broker, speller->speller);
      g_free(speller);
    }
}

static Request *
gt_language_detection_request_new (RequestType type)
{
  Request *request;

  request = g_new0(Request, 1);
  request->type = type;

  return request;
}

static void
gt_language_detection_request_free (Request *request)
{
  g_return_if_fail(request != NULL);

  g_slist_foreach(request->spellers, (GFunc) gt_language_detection_speller_unref, NULL);
  g_slist_free(request->spellers);
  g_slist_foreach(request->words, (GFunc) g_free, NULL);
  g_slist_free(request->words);
  g_free(request);
}

void
gt_language_detection_detect_language (const char *text)
{
  g_return_if_fail(text != NULL);

  G_LOCK(self);

  if (detect_queue)
    {
      char **words;
      GSList *used_words = NULL;
      int i;
      
      words = g_strsplit(text, " ", 0);

      for (i = 0; words[i]; i++)
	{
	  g_strstrip(words[i]);
	  if (gt_language_detection_use_word(words[i]))
	    used_words = g_slist_append(used_words, words[i]);
	  else
	    g_free(words[i]);
	}

      g_free(words);

      if (used_words)
	{
	  Request *request;

	  request = gt_language_detection_request_new(REQUEST_DETECT_LANGUAGE);
	  request->spellers = gt_g_slist_copy_deep(spellers, (GTCopyFunc) gt_language_detection_speller_ref, NULL);
	  request->words = used_words;
	  
	  g_async_queue_push(detect_queue, request);
	}
    }

  G_UNLOCK(self);
}

static gpointer
gt_language_detection_detect_thread (gpointer data)
{
  GAsyncQueue *queue = data;
  Request *request;
  gboolean disable = FALSE;

  do
    {
      request = g_async_queue_pop(queue);

    start:
      if (request->type == REQUEST_DETECT_LANGUAGE && request->spellers)
	{
	  GSList *l;
	  Speller *speller1;
	  Speller *speller2;
	  
	  GT_LIST_FOREACH(l, request->spellers)
	    {
	      Speller *speller = l->data;
	      speller->score = 0;
	    }
	  
	  GT_LIST_FOREACH(l, request->words)
	    {
	      const char *word = l->data;
	      GSList *m;

	      GT_LIST_FOREACH(m, request->spellers)
	        {
		  Speller *speller = m->data;
		  Request *new_request;
	      
		  if ((new_request = g_async_queue_try_pop(detect_queue)))
		    {
		      gt_language_detection_request_free(request);
		      request = new_request;
		      goto start;
		    }

		  if (enchant_dict_check(speller->speller, word, strlen(word)) == 0)
		    speller->score++;
		}
      
	      request->spellers = g_slist_reverse(g_slist_sort(request->spellers, (GCompareFunc) gt_language_detection_speller_compare_score));
	
	      speller1 = g_slist_nth_data(request->spellers, 0);
	      speller2 = g_slist_nth_data(request->spellers, 1);
	      
	      /*
	       * If the first language leads by 10 points or more, we
	       * assume it is going to win, and we stop here for
	       * performance.
	       */
	      if (speller1->score - speller2->score >= 10)
		break;
	    }
      
	  /* we only give an opinion if there is no tie */
	  if (speller1->score != speller2->score)
	    result_func(speller1->tag, result_user_data);
	}

      disable = request->type == REQUEST_DISABLE;
      gt_language_detection_request_free(request);
    }
  while (! disable);
  
  g_async_queue_unref(queue);

  return NULL;
}

static gboolean
gt_language_detection_use_word (const char *word)
{
  g_return_val_if_fail(word != NULL, FALSE);
  
  for (; *word; word = g_utf8_next_char(word))
    if (g_unichar_isalpha(g_utf8_get_char(word)))
      return TRUE;
  
  return FALSE;
}

static int
gt_language_detection_speller_compare_score (const Speller *a, const Speller *b)
{
  return a->score - b->score;
}
